unit Ibase_level2_server;

interface

uses Ibase_level1_server,ibase,syncobjs,controls,classes;

type
 TConnectPlus =class;

//--------------------------------------------
//         
//--------------------------------------------
 TTypeField_ =(tfString,tfInteger,tfDate,tfBlob,tfDouble,tfTimestamp,tfDateNew);
 TField_ =class(TObject)
  protected
   FAlternativeName : TStringList; //  (  View)
   FNameField : string; // 
   FNameAlias : string; // 
   FIndIsNull      : smallint;      //    Null
   FValueInteger   : integer;       //   tfInteger
   FValueString    : PChar;         //   tfString
   FValueDouble    : double;        //   tfDouble
   FValueDate      : ISC_QUAD;      //   tfDate
   FValueBlob      : ISC_QUAD;      //   tfBlob
   FValueTimestamp : ISC_TIMESTAMP; //   tfTimeStamp
   FValueDateNew   : ISC_DATE;      //   tfDateNew
   function GetIsNull: boolean; virtual;
   procedure SetIsNull(const Value: boolean); virtual;
   function GetType : TTypeField_; virtual; abstract; //  
   function BelongName(n : string; flagAlias : boolean) : boolean;
   function Set_XSQLVAR(p : Tpointer_XSQLVAR) : boolean; virtual;
  public
   procedure SetAlias(a : string); // 

   property IsNull : boolean read GetIsNull write SetIsNull;
   property NameField : string read FNameField;
   property NameAlias : string read FNameAlias write SetAlias;
   
   constructor Create(NameField : string);
   destructor Destroy; override;
   procedure AddAlternativeName(n : string);
   procedure Assign(f : TField_); virtual;
 end;

// Integer
 TFieldInteger =class(TField_)
  protected
   function GetType : TTypeField_; override;
   function GetValue: integer; virtual;
   procedure SetValue(const Value: integer); virtual;
  public
   property Value : integer read GetValue write SetValue;
   constructor Create(NameField : string);
 end;

// Double
 TFieldDouble =class(TField_)
  protected
   function GetType : TTypeField_; override;
   function GetValue: double;
   procedure SetValue(const Value: double);
  public
   property Value : double read GetValue write SetValue;
   constructor Create(NameField : string);
 end;

// String
 TFieldString =class(TField_)
  protected
   FLength : integer;
   FAlwaysUpper : boolean; // true   SetValue   AnsiUpperCase
   function GetType : TTypeField_; override;
   function GetValue: string; virtual;
   procedure SetValue(const Value: string); virtual;
  public
   property Value : string read GetValue write SetValue;
   property Length : integer read FLength;
   constructor Create(NameField : string; length_ : integer);
   destructor Destroy; override;
   procedure SetAlwaysUpper; // FAlwaysUpper =true
   procedure Assign(f : TField_); override;
 end;

// Date
 TFieldDate =class(TField_)
  protected
   function GetType : TTypeField_; override;
   function GetDate: TDate;
   procedure SetDate(const Value: TDate);
  public
   property Value : TDate read GetDate write SetDate;
   constructor Create(NameField : string);
 end;

// TimeStamp
 TFieldTimeStamp =class(TField_)
  protected
   function GetType : TTypeField_; override;
   function GetDate: TDateTime;
   procedure SetDate(const Value: TDateTime);
  public
   property Value : TDateTime read GetDate write SetDate;
   constructor Create(NameField : string);
 end;

// Blob (ID Blob)
 TFieldBlob =class(TField_)
  protected
   function GetType : TTypeField_; override;
   function GetData: ISC_QUAD;
   procedure SetData(const Value: ISC_QUAD);
  public
   property Value : ISC_QUAD read GetData write SetData;
   constructor Create(NameField : string);
 end;

//--------------------------------------------
//  
//--------------------------------------------
 TRecord_ = class(TObject)
  protected
   FList : TList;
   FListInnerRecords : TList;
{   XSQLVAR   
    .  true    
 false .        -
      aliasa.}
   function Set_XSQLVAR(p : Tpointer_XSQLVAR) : boolean; virtual;
//    n.      nil
   function GetField(n : string) : TField_;
{   t   n.  t = tfString l - ,
   .}
   function AddField(t : TTypeField_; n : string; l : integer = 0) : TField_;
{  r  FListInnerRecord. FList   
  r.Flist.Items}
   procedure AddInnerRecord(r : TRecord_);

   function GetCountFields: integer;
   function GetFields(index: integer): TField_;
  public
   property CountFields : integer read GetCountFields;
   property Fields[index : integer] : TField_ read GetFields;
   constructor Create;
   destructor Destroy; override;
//  Source
   procedure Assign(Source : TRecord_); virtual;
//    null
   procedure SetFieldsToNull;
// 
   function Add(t : TField_) : TField_;
// 
   procedure Delete(t : TField_);
 end;

//--------------------------------------------
// 
//--------------------------------------------
//   TReadWriteTPB
 TReadWriteTransaction =class(TTransaction)
  public
   constructor Create(connect : TConnection; no_rec : boolean =false);
   destructor Destroy; override;
 end;

//   TReadOnlyTPB
 TReadOnlyTransaction =class(TTransaction)
  public
   constructor Create(connect : TConnection);
   destructor Destroy; override;
 end;

 TConcurrencyOnlyReadTransaction =class(TTransaction)
  public
   constructor Create(connect : TConnection);
   destructor Destroy; override;
 end;

//--------------------------------------------
// Select 
//--------------------------------------------
 TSelect_ =class(TDSQL_metod3)
  protected
   FRecord_ : TRecord_; //  
   FTransaction : TTransaction;
   FConnect     : TConnection;
   FIsOutTransaction : boolean;
   function GetStrSelect : string; virtual; abstract;
  public
   property pRecord     : TRecord_     read FRecord_;
   property Transaction : TTransaction read FTransaction;
   constructor Create(connect : TConnection; t : TTransaction = nil);
   destructor Destroy; override;
 end;

 TSelectWithParam_ =class(TDSQL_metod4)
  protected
   FRecord_ : TRecord_; //  
   FTransaction : TTransaction;
   FConnect     : TConnection;
   FIsOutTransaction : boolean;
   function GetStrSelect : string; virtual; abstract;
// GetParams   
   procedure GetParams(p : array of TField_); virtual;
  public
   property pRecord : TRecord_ read FRecord_;
   constructor Create(connect : TConnection; t : TTransaction = nil; isParametr : boolean = true);
   destructor Destroy; override;
 end;

 TTable_ =class(TObject)
  protected
   FConnect: TConnectPlus;
   procedure Update_(strSQL : string; t : TTransaction; f : array of TField_);
   procedure Delete_(nameTable : string; t : TTransaction; IDs : array of integer);
   procedure Update2_(const nameTable : string; r : TRecord_; ID : integer; t : TTransaction);
  public
   constructor Create(connect : TConnectPlus);
 end;

//-------------------------------------------
//    
//-------------------------------------------
 TConnectPlus =class(TConnection)
   protected
    FIB_Events : TIB_Events;
    FUser_ : string;
    FRole_ : string;
    FPath_ : string;
   public
    property IB_Events : TIB_Events read FIB_Events;
    property User_ : string read FUser_;
    property Role_ : string read FRole_;
    property Path_ : string read FPath_;
    constructor Create;
    destructor Destroy; override;
    procedure Detach; override;
//  
    procedure Attach(base_,user,password : string; CharSet : string ='WIN1251'; role : string = '');
// ReadWrite
    function cReadWriteTransaction(no_rec : boolean =false) : TReadWriteTransaction;
// ReadOnly
    function cReadOnlyTransaction : TReadOnlyTransaction;
// ConcurrencyReadOnly
    function cConcurrencyOnlyReadTransaction : TConcurrencyOnlyReadTransaction;
//      
    function IncGenerator(n : string) : integer;
//       
    procedure ExecProcInOut(strSQL : string; t : TTransaction; in_ : array of TField_; out_ : array of TField_);
// SQL      
    procedure ExecSQL(t : TTransaction; strSQL : string);
// Blob     list_.
    procedure Read_Blob(list_ : TStrings; id_blob : ISC_QUAD; t : TTransaction = nil); overload;
    function  Read_Blob(id_blob : ISC_QUAD; t : TTransaction = nil) : string; overload;
// list_  Blob 
    function Write_Blob(list_ : TStrings; t : TTransaction) : ISC_QUAD; overload;
    function Write_Blob(s : string; t : TTransaction) : ISC_QUAD; overload;
//  SELECT COUNT(*)   
    function GetCountRecords(table_ : string; where_ : string ='') : integer;
//   
    procedure Insert_(nameTable : string; t : TTransaction; r : TRecord_);
//   
    procedure Update_(strSQL : string; t : TTransaction; r: TRecord_);
 end;

//     XSQLVAR
function GetNameColumn(p : Tpointer_XSQLVAR) : string;
//     XSQLVAR
function GetNameAlias(p : Tpointer_XSQLVAR) : string;

var
{     Float   String,
      DecimalSeparator.}
 SectionAccessToDecimalSeparator : TCriticalSection;

implementation

uses SysUtils,ibase_exception;

type
//-------------------------------------------
// TTPB
//-------------------------------------------
 TReadOnlyTPB =class(TTPB)
   public
    constructor Create;
 end;

 TReadWriteTPB =class(TTPB)
   public
    constructor Create(no_rec : boolean = false);
 end;

 TConcurrencyOnlyReadTPB =class(TTPB)
   public
    constructor Create;
 end;

//   
//----------------------------------
 TRecordForID =class(TRecord_)
  private
   function GetID: integer;
  public
   property ID : integer read GetID;
   constructor Create;
 end;

 TSelectNextID =class(TSelect_)
  private
   FNameGenerator : string;
   function GetID : integer;
   function GetStrSelect : string; override;
  public
   property ID : integer read GetID;
   constructor Create(connect : TConnection; NameGen : string);
 end;

//    
//-----------------------------------------
 TRecordCountRecords =class(TRecord_)
  private
   function GetCount: integer;
  public
   property count : integer read GetCount;
   constructor Create;
 end;

 TSelectCountRecords =class(TSelect_)
  protected
   FNameTable : string;
   FWhere : string;
   function GetRecord : TRecordCountRecords;
   function GetStrSelect : string; override;
  public
   property Record_ : TRecordCountRecords read GetRecord;
   constructor Create(connect : TConnection; table_ : string; where_ : string ='');
 end;

function GetNameColumn(p: Tpointer_XSQLVAR): string;
begin
 Result :=AnsiUpperCase(Copy(p^.sqlname,0,p^.sqlname_length));
end;

function GetNameAlias(p: Tpointer_XSQLVAR): string;
begin
 Result :=AnsiUpperCase(Copy(p^.aliasname,0,p^.aliasname_length));
end;

{ TConnectPlus }

procedure TConnectPlus.Attach(base_, user, password, CharSet, role: string);
 var ParametrDPB : TParametrDPB;
     DPB : TDPB;
 begin
  ParametrDPB := TParametrDPB.Create;
  DPB := TDPB.Create;
  try
   ParametrDPB.user :=user;
   ParametrDPB.password :=password;
   ParametrDPB.CharacterSet :=CharSet;
   ParametrDPB.role :=role;
   DPB.FillDPB(ParametrDPB);
   inherited Attach(DPB,base_);

   FUser_ :=AnsiUpperCase(ParametrDPB.user);
   FRole_ :=AnsiUpperCase(ParametrDPB.role);
   FPath_ :=base_;
  finally
   ParametrDPB.Free;
   DPB.Free;
  end;
end;

function TConnectPlus.cReadOnlyTransaction: TReadOnlyTransaction;
begin
 Result :=TReadOnlyTransaction.Create(self);
end;

function TConnectPlus.cReadWriteTransaction(no_rec : boolean =false) : TReadWriteTransaction;
begin
 Result :=TReadWriteTransaction.Create(self,no_rec);
end;

constructor TConnectPlus.Create;
begin
 inherited Create;
 FIB_Events :=TIB_Events.Create(self);
 FUser_ :=''; FRole_ :='';
end;

destructor TConnectPlus.Destroy;
begin
 FIB_Events.Free;
 inherited;
end;

procedure TConnectPlus.Detach;
begin
 if IB_Events.FlagRegistration then IB_Events.UnRegistration;
 inherited Detach;
 FUser_ :=''; FRole_ :='';
end;

procedure TConnectPlus.ExecProcInOut(strSQL : string; t: TTransaction; in_, out_: array of TField_);
var p : TDSQL_metod4;
    i : integer;
begin
 p :=TDSQL_metod4.Create;
 try
  p.Prepare(self,t,strSQL);

  if p.CountParametrs <> length(in_) then raise EIbaseError.Create(' TConnectPlus.  ExecProcInOut.   in_      .');
  for i :=0 to p.CountParametrs -1 do begin
    StrCopy(p.Parametr[i]^.sqlname,PChar(in_[i].NameField));
    p.Parametr[i]^.sqlname_length :=length(in_[i].NameField);
    in_[i].Set_XSQLVAR(p.Parametr[i]);
  end;

  if p.CountFields <> length(out_) then raise EIbaseError.Create(' TConnectPlus.  ExecProcInOut.   out_      .');
  for i :=0 to p.CountFields -1 do begin
    StrCopy(p.Field[i]^.sqlname,PChar(out_[i].NameField));
    p.Field[i]^.sqlname_length :=length(out_[i].NameField);
    out_[i].Set_XSQLVAR(p.Field[i]);
  end;

  p.ExecuteProc;
 finally
  p.Free;
 end;
end;

procedure TConnectPlus.ExecSQL(t: TTransaction; strSQL : string);
begin
 Execute_metod1(self,t,strSQL);
end;

function TConnectPlus.GetCountRecords(table_: string; where_ : string =''): integer;
var s : TSelectCountRecords;
begin
 s :=TSelectCountRecords.Create(self,table_,where_);
 try
  s.Open_cursor;
  s.Fetch;
  Result :=s.Record_.count;
 finally
  s.Free;
 end;
end;

function TConnectPlus.IncGenerator(n: string): integer;
var s : TSelectNextID;
begin
 Result :=0;
 s :=TSelectNextID.Create(self,n);
 try
  s.Open_cursor;
  s.Fetch;
  if not s.eof then Result :=s.ID;
 finally
  s.Free;
 end;
end;

procedure TConnectPlus.Read_Blob(list_: TStrings; id_blob: ISC_QUAD; t : TTransaction = nil);
 var Transaction : TTransaction;
     Text_Blob   : TText_Blob;
 begin
  if t = nil then Transaction :=cReadOnlyTransaction
             else Transaction :=t;
  try
    Text_Blob :=TText_Blob.Create(self.db_handle,Transaction.tr_handle);
    try
     Text_Blob.id_blob :=id_blob;
     Text_Blob.Read_Blob;
     list_.Assign(Text_Blob);
    finally
     Text_Blob.Free;
    end;
  finally
   if t = nil then Transaction.Free;
  end;
end;

procedure TConnectPlus.Insert_(nameTable: string; t: TTransaction; r: TRecord_);
var objectInsert : TDSQL_metod2;
    i : integer;
    strSQL : string;
begin
 objectInsert :=TDSQL_metod2.Create;
 try
  strSQL :='INSERT INTO ' + nameTable + '(';
  for i :=0 to r.CountFields -1 do begin
    strSQL :=strSQL + r.Fields[i].NameField;
    if i < r.CountFields -1 then strSQL :=strSQL + ',';
  end;
  strSQL :=strSQL + ') VALUES (';
  for i :=0 to r.CountFields -1 do begin
    strSQL :=strSQL + '?';
    if i < r.CountFields -1 then strSQL :=strSQL + ',';
  end;
  strSQL :=strSQL + ')';

  objectInsert.Prepare(self,t,strSQL);

  for i :=0 to objectInsert.CountParametr -1 do begin
    StrCopy(objectInsert.Parametr[i]^.sqlname,PChar(r.Fields[i].NameField));
    objectInsert.Parametr[i]^.sqlname_length :=length(r.Fields[i].NameField);
    if r.Fields[i].GetType = tfString then
      objectInsert.Parametr[i]^.sqllen :=length(r.Fields[i].FValueString);
    r.Fields[i].Set_XSQLVAR(objectInsert.Parametr[i]);
  end;

  objectInsert.Execute;
 finally
  objectInsert.Free;
 end;
end;

function TConnectPlus.Read_Blob(id_blob: ISC_QUAD; t : TTransaction = nil): string;
var l : TStringList;
begin
 l :=TStringList.Create;
 try
  Read_Blob(l,id_blob,t);
  Result :=l.Text;
 finally
  l.Free;
 end;
end;

function TConnectPlus.Write_Blob(list_: TStrings; t: TTransaction): ISC_QUAD;
 var Text_Blob   : TText_Blob;
 begin
  Text_Blob :=TText_Blob.Create(self.db_handle,t.tr_handle);
  try
   Text_Blob.Assign(list_);
   Text_Blob.Write_Blob;
   Result :=Text_Blob.id_blob;
  finally
   Text_Blob.Free;
  end;
end;

function TConnectPlus.Write_Blob(s: string; t: TTransaction): ISC_QUAD;
var l : TStringList;
begin
 l :=TStringList.Create;
 try
  l.Text :=s;
  Result :=Write_Blob(l,t);
 finally
  l.Free;
 end;
end;

procedure TConnectPlus.Update_(strSQL: string; t: TTransaction; r: TRecord_);
var objectUpdate : TDSQL_metod2;
    i : integer;
begin
 objectUpdate :=TDSQL_metod2.Create;
 try
  objectUpdate.Prepare(self,t,strSQL);

  for i :=0 to objectUpdate.CountParametr -1 do begin
    StrCopy(objectUpdate.Parametr[i]^.sqlname,PChar(r.Fields[i].NameField));
    objectUpdate.Parametr[i]^.sqlname_length :=length(r.Fields[i].NameField);
    if r.Fields[i].GetType = tfString then
      objectUpdate.Parametr[i]^.sqllen :=length(r.Fields[i].FValueString);
    r.Fields[i].Set_XSQLVAR(objectUpdate.Parametr[i]);
  end;

  objectUpdate.Execute;
 finally
  objectUpdate.Free;
 end;
end;

function TConnectPlus.cConcurrencyOnlyReadTransaction: TConcurrencyOnlyReadTransaction;
begin
 Result :=TConcurrencyOnlyReadTransaction.Create(self);
end;

{ TReadOnlyTPB }

constructor TReadOnlyTPB.Create;
 var ParametrTPB : TParametrTPB;
 begin
  inherited Create;
  ParametrTPB :=TParametrTPB.Create;
  try
    ParametrTPB.Access_mode :=tpb_read;                   //   
    ParametrTPB.Isolation_level :=tpb_read_committed_rec; //         commit,       
    ParametrTPB.Lock_resolution :=tpb_nowait;             //   
    FillTPB(ParametrTPB);
  finally
   ParametrTPB.Free;
  end;
end;

{ TReadWriteTPB }

constructor TReadWriteTPB.Create(no_rec : boolean = false);
 var ParametrTPB : TParametrTPB;
 begin
  inherited Create;
  ParametrTPB :=TParametrTPB.Create;
  try
   ParametrTPB.Access_mode :=tpb_write; //    
   if no_rec
//         commit,        lock error
     then ParametrTPB.Isolation_level :=tpb_read_committed_no_rec
//         commit,       
     else ParametrTPB.Isolation_level :=tpb_read_committed_rec;
   ParametrTPB.Lock_resolution :=tpb_nowait; //   
   FillTPB(ParametrTPB);
  finally
   ParametrTPB.Free;
  end;
end;

{ TReadWriteTransaction }

constructor TReadWriteTransaction.Create(connect: TConnection; no_rec : boolean =false);
var ReadWriteTPB : TReadWriteTPB;
begin
 inherited Create;
 ReadWriteTPB :=TReadWriteTPB.Create(no_rec);
 try
   Start_transaction(connect,ReadWriteTPB);
 finally
  ReadWriteTPB.Free;
 end;
end;

destructor TReadWriteTransaction.Destroy;
begin
  if IsActiveTransaction then Commit;
  inherited;
end;

{ TReadOnlyTransaction }

constructor TReadOnlyTransaction.Create(connect: TConnection);
var ReadOnlyTPB : TReadOnlyTPB;
begin
 inherited Create;
 ReadOnlyTPB :=TReadOnlyTPB.Create;
 try
   Start_transaction(connect,ReadOnlyTPB);
 finally
   ReadOnlyTPB.Free;
 end;
end;

destructor TReadOnlyTransaction.Destroy;
begin
 if IsActiveTransaction then Commit;
 inherited;
end;

{ TSelect_ }

constructor TSelect_.Create(connect: TConnection; t: TTransaction);
var TPB : TReadOnlyTPB;
    i : integer;
begin
 FTransaction :=nil; //       
 inherited Create;

//    
 if t = nil then begin
   FIsOutTransaction :=false;
   TPB :=TReadOnlyTPB.Create;
   try
     FTransaction :=TTransaction.Create;
     FTransaction.Start_transaction(connect,TPB);
   finally
    TPB.Free;
   end;
// 
  end else begin
    FIsOutTransaction :=true;
    FTransaction :=t;
  end;
  FConnect :=connect;

//  SQL     
  Prepare(connect,FTransaction,GetStrSelect);
  for i :=0 to GetCountFields-1 do
    if not FRecord_.Set_XSQLVAR(Field[i]) then
      raise EIbaseError.Create(' ' + GetNameColumn(Field[i]) + '  .    , , .');
end;

destructor TSelect_.Destroy;
begin
 if IsOpen then Close_cursor;
 if (FTransaction <> nil) and not FIsOutTransaction then begin
   if FTransaction.IsActiveTransaction then FTransaction.Commit;
   FTransaction.Free;
 end;
 FRecord_.Free;
 inherited Destroy;
end;

{ TField_ }

function TField_.GetIsNull: boolean;
begin
 Result :=(FIndIsNull = VALUE_IS_NULL);
end;

function TField_.Set_XSQLVAR(p: Tpointer_XSQLVAR): boolean;
begin
 Result :=true;
 if ((GetNameAlias(p) <> GetNameColumn(p)) and (NameAlias <> '') and BelongName(GetNameAlias(p),true)) or
    (((GetNameAlias(p) = '') or (GetNameAlias(p) = GetNameColumn(p))) and BelongName(GetNameColumn(p),false)) or
//        
     ((GetNameAlias(p) <> GetNameColumn(p)) and (AnsiUpperCase(GetNameAlias(p)) = AnsiUpperCase(NameField))) then begin
   case GetType of
    tfInteger   : begin p^.sqldata :=@FValueInteger;   p^.sqltype :=SQL_LONG+1;       end;
    tfDouble    : begin p^.sqldata :=@FValueDouble;    p^.sqltype :=SQL_DOUBLE+1      end;
    tfString    : begin p^.sqldata :=FValueString;     p^.sqltype :=SQL_TEXT+1;       end;
    tfDate      : begin p^.sqldata :=@FValueDate;      p^.sqltype :=SQL_DATE+1;       end;
    tfBlob      : begin p^.sqldata :=@FValueBlob;      p^.sqltype :=SQL_BLOB+1;       end;
    tfTimestamp : begin p^.sqldata :=@FValueTimestamp; p^.sqltype :=SQL_TIMESTAMP +1; end;
    tfDateNew   : begin p^.sqldata :=@FValueDateNew;   p^.sqltype :=SQL_TYPE_DATE +1; end;
   end;
   p^.sqlind :=@FIndIsNull;
 end else Result :=false;
end;

procedure TField_.SetIsNull(const Value: boolean);
begin
 if Value then FIndIsNull :=VALUE_IS_NULL
          else FIndIsNull :=VALUE_NOT_NULL;
end;

constructor TField_.Create(NameField: string);
begin
 inherited Create;
 FAlternativeName :=TStringList.Create;

 if length(NameField) > maxForXSQLVAR then raise EIbaseError.Create(' TField_.  Create.      maxForXSQLVAR');
 FNameField :=NameField;
 FNameAlias :='';
 FValueString :=nil;
end;

procedure TField_.SetAlias(a: string);
begin
 FNameAlias :=a;
end;

destructor TField_.Destroy;
begin
 FAlternativeName.Free;
 inherited Destroy;
end;

procedure TField_.AddAlternativeName(n: string);
begin
 FAlternativeName.Add(n);
end;

function TField_.BelongName(n: string; flagAlias : boolean): boolean;
var i : integer;
begin
 n :=AnsiUpperCase(n);
 if flagAlias
  then Result :=AnsiUpperCase(NameAlias) = n
  else Result :=AnsiUpperCase(NameField) = n;
 if not Result then
   for i :=0 to FAlternativeName.Count -1 do
    if n = AnsiUpperCase(FAlternativeName.Strings[i]) then begin
      Result :=true;
      break;
    end;
end;

procedure TField_.Assign(f: TField_);
begin
 FIndIsNull      :=f.FIndIsNull;
 FValueInteger   :=f.FValueInteger;
 FValueDouble    :=f.FValueDouble;
 FValueDate      :=f.FValueDate;
 FValueBlob      :=f.FValueBlob;
 FValueTimestamp :=f.FValueTimestamp;
 FValueDateNew   :=f.FValueDateNew;
 if f.FValueString = nil then FValueString :=nil
                         else StrCopy(FValueString,f.FValueString);
end;

{ TFieldInteger }

constructor TFieldInteger.Create(NameField: string);
begin
 inherited Create(NameField);
 FValueInteger :=0; // 
end;

function TFieldInteger.GetType: TTypeField_;
begin
 Result :=tfInteger;
end;

function TFieldInteger.GetValue: integer;
begin
 Result :=FValueInteger;
end;

procedure TFieldInteger.SetValue(const Value: integer);
begin
 FValueInteger :=Value; IsNull :=false;
end;

{ TFieldDouble }

constructor TFieldDouble.Create(NameField: string);
begin
 inherited Create(NameField);
 Value :=0; // 
end;

function TFieldDouble.GetType: TTypeField_;
begin
 Result :=tfDouble;
end;

function TFieldDouble.GetValue: double;
begin
 Result :=FValueDouble;
end;

procedure TFieldDouble.SetValue(const Value: double);
begin
 FValueDouble :=Value; IsNull :=false;
end;

{ TFieldString }

procedure TFieldString.Assign(f: TField_);
begin
 if f is TFieldString then begin
   FLength      :=(f as TFieldString).FLength;
   FAlwaysUpper :=(f as TFieldString).FAlwaysUpper;
 end;
 inherited Assign(f);
end;

constructor TFieldString.Create(NameField: string; length_: integer);
begin
 inherited Create(NameField);
 FLength :=length_;
 FAlwaysUpper :=false;
 FValueString :=StrAlloc(FLength+1); FValueString[FLength] :=#0;
 Value :=''; // 
end;

destructor TFieldString.Destroy;
begin
 StrDispose(FValueString);
 inherited Destroy;;
end;

function TFieldString.GetType: TTypeField_;
begin
 Result :=tfString;
end;

function TFieldString.GetValue: string;
begin
 if isNull then Result :=''
           else Result :=trim(FValueString);
end;

procedure TFieldString.SetAlwaysUpper;
begin
 FAlwaysUpper :=true;
end;

procedure TFieldString.SetValue(const Value: string);
var v : string;
begin
 v :=trim(Value);
 if FAlwaysUpper then v :=AnsiUpperCase(v);
 StrCopy(FValueString,PChar(Copy(v,1,FLength)));
 IsNull :=v = ''; 
end;

{ TFieldDate }

constructor TFieldDate.Create(NameField: string);
begin
 inherited Create(NameField);
 Value :=0; // 
end;

function TFieldDate.GetDate: TDate;
begin
 if Interbase_dialect = 3 then Result :=ibDateOnlyTo(@FValueDateNew)
                          else Result :=ibDateTo(@FValueDate);
end;

function TFieldDate.GetType: TTypeField_;
begin
 if Interbase_dialect =3 then Result :=tfDateNew
                         else Result :=tfDate;
end;

procedure TFieldDate.SetDate(const Value: TDate);
begin
 if Interbase_dialect =3 then ToibDateOnly(@FValueDateNew,Value)
                         else ToibDate(@FValueDate,Value);
 isNull :=Value = 0;
end;

{ TFieldBlob }

constructor TFieldBlob.Create(NameField: string);
begin
 inherited Create(NameField);
 IsNull :=true; // 
end;

function TFieldBlob.GetData: ISC_QUAD;
begin
 Result :=FValueBlob;
end;

function TFieldBlob.GetType: TTypeField_;
begin
 Result :=tfBlob;
end;

procedure TFieldBlob.SetData(const Value: ISC_QUAD);
begin
 FValueBlob :=Value; IsNull :=false;
end;

{ TRecord_ }

function TRecord_.Add(t: TField_) : TField_;
begin
 FList.Add(t);
 Result :=t;
end;

function TRecord_.AddField(t: TTypeField_; n: string; l: integer) : TField_;
var i : integer;
begin
 i :=-1; //   
 case t of
  tfString         : i :=FList.Add(TFieldString.Create(n,l));
  tfInteger        : i :=FList.Add(TFieldInteger.Create(n));
  tfDate,tfDateNew : i :=FList.Add(TFieldDate.Create(n));
  tfBlob           : i :=FList.Add(TFieldBlob.Create(n));
  tfDouble         : i :=FList.Add(TFieldDouble.Create(n));
  tfTimeStamp      : i :=FList.Add(TFieldTimeStamp.Create(n));
 end;
 Result :=TObject(FList.Items[i]) as TField_;
end;

procedure TRecord_.AddInnerRecord(r: TRecord_);
var i : integer;
begin
 FListInnerRecords.Add(r);
 for i :=0 to r.FList.Count -1 do FList.Add(r.FList.Items[i]);
end;

procedure TRecord_.Assign(Source: TRecord_);
// true     
 function CompareAlternativeName(f1,f2 : TField_) : boolean;
  var i : integer;
  begin
   Result :=false;
   if f1.FAlternativeName.Count = f2.FAlternativeName.Count then begin
     Result :=true;
     for i :=0 to f1.FAlternativeName.Count-1 do
      if f1.FAlternativeName.Strings[i] <> f2.FAlternativeName.Strings[i] then begin
        Result :=false;
        break;
      end;
   end;
  end;

var i,j : integer;
begin
 for i :=0 to FList.Count -1 do
  with (TObject(FList.Items[i]) as TField_) do begin
    for j :=0 to Source.FList.Count -1 do
      if (NameField = TField_(Source.FList.Items[j]).NameField) and
         (NameAlias = TField_(Source.FList.Items[j]).NameAlias) and
         CompareAlternativeName(TObject(self.FList.Items[i]) as TField_,TField_(Source.FList.Items[j])) and
         (GetType = TField_(Source.FList.Items[j]).GetType) then begin
        Assign(TField_(Source.FList.Items[j]));
        break;
      end;
  end;
end;

constructor TRecord_.Create;
begin
 inherited Create;
 FList :=TList.Create;
 FListInnerRecords :=TList.Create;
end;

procedure TRecord_.Delete(t: TField_);
var i : integer;
begin
 i :=FList.IndexOf(t);
 FList.Delete(i);
 t.Free;
end;

destructor TRecord_.Destroy;
var i,j : integer;
begin
//   TRecord_     
  for i :=0 to FListInnerRecords.Count -1 do begin
    for j :=0 to TRecord_(FListInnerRecords.Items[i]).FList.Count -1 do
      FList.Delete(FList.IndexOf(TRecord_(FListInnerRecords.Items[i]).FList.Items[j]));
    TObject(FListInnerRecords.Items[i]).Free;
  end;
//  
  for i :=0 to FList.Count -1 do TObject(FList.Items[i]).Free;
  
  FList.Free;
  FListInnerRecords.Free;
  inherited Destroy;
end;

function TRecord_.GetCountFields: integer;
begin
 Result :=FList.Count;
end;

function TRecord_.GetField(n: string): TField_;
var i : integer;
begin
 Result :=nil;
 for i :=0 to FList.Count -1 do
   if (TObject(FList.Items[i]) as TField_).BelongName(n,false) then begin
     Result :=FList.Items[i];
     break;
   end;
end;

function TRecord_.GetFields(index: integer): TField_;
begin
 Result :=TObject(FList.Items[index]) as TField_;
end;

procedure TRecord_.SetFieldsToNull;
var i : integer;
begin
 for i :=0 to FList.Count -1 do
  (TObject(FList.Items[i]) as TField_).IsNull :=true;
end;

function TRecord_.Set_XSQLVAR(p: Tpointer_XSQLVAR): boolean;
var i : integer;
begin
 Result :=false;
 for i :=0 to FList.Count -1 do begin
  Result :=(TObject(FList[i]) as TField_).Set_XSQLVAR(p);
  if Result then break;
 end;
end;

{ TTable_ }

constructor TTable_.Create(connect: TConnectPlus);
begin
 inherited Create;
 FConnect :=connect;
end;

procedure TTable_.Delete_(nameTable: string; t: TTransaction; IDs: array of integer);
var objectDel : TDSQL_metod1;
    s,strSQL : string;
    i : integer;
begin
 if length(IDs) = 0 then raise EIbaseError.Create(' TTable_.  Delete_.  .');

 strSQL :='DELETE FROM ?Table WHERE ID IN (?IDs)';
 strSQL :=StringReplace(strSQL,'?Table',nameTable,[rfIgnoreCase]);

 objectDel :=TDSQL_metod1.Create;
 try
  s :='';
  for i :=0 to High(IDs) do begin
    if s <> '' then s :=s + ',';
    s :=s+IntToStr(IDs[i]);
  end;
  strSQL :=StringReplace(strSQL,'?IDs',s,[]);
  objectDel.Prepare(FConnect,t,strSQL);
  objectDel.Execute;
 finally
  objectDel.Free;
 end;
end;

procedure TTable_.Update2_(const nameTable: string; r: TRecord_; ID: integer; t : TTransaction);
var i : integer;
    s : string;
begin
 s :='UPDATE ' + nameTable + ' SET ';
 for i :=0 to r.CountFields -1 do begin
   s :=s + r.Fields[i].NameField + ' = ?';
   if i < r.CountFields -1 then s :=s + ', ';
 end;
 s :=s + ' WHERE ID =' + IntToStr(ID);

 FConnect.Update_(s,t,r);
end;

procedure TTable_.Update_(strSQL: string; t: TTransaction; f: array of TField_);
var objectUpdate : TDSQL_metod2;
    i : integer;
begin
 objectUpdate :=TDSQL_metod2.Create;
 try
  objectUpdate.Prepare(FConnect,t,strSQL);

  if objectUpdate.CountParametr <> length(f) then raise EIbaseError.Create(' TTable_.  Update_.       .');
  for i :=0 to objectUpdate.CountParametr -1 do begin
    StrCopy(objectUpdate.Parametr[i]^.sqlname,PChar(f[i].NameField));
    objectUpdate.Parametr[i]^.sqlname_length :=length(f[i].NameField);
    if f[i].GetType = tfString then
      objectUpdate.Parametr[i]^.sqllen :=length(f[i].FValueString);
    f[i].Set_XSQLVAR(objectUpdate.Parametr[i]);
  end;

  objectUpdate.Execute;
 finally
  objectUpdate.Free;
 end;
end;

{ TRecordForID }

constructor TRecordForID.Create;
begin
 inherited Create;
 AddField(tfInteger,'ID').SetAlias('ID');
end;

function TRecordForID.GetID: integer;
begin
  Result :=(GetField('ID') as TFieldInteger).Value;
end;

{ TSelectNextID }

constructor TSelectNextID.Create(connect: TConnection; NameGen: string);
begin
 FRecord_ :=TRecordForID.Create;
 FNameGenerator :=NameGen;
 inherited Create(connect,nil);
end;

function TSelectNextID.GetID: integer;
begin
 Result :=(FRecord_ as TRecordForID).ID;
end;

function TSelectNextID.GetStrSelect: string;
begin
 Result :='SELECT CAST(GEN_ID(?,1) AS INTEGER) AS ID FROM RDB$GENERATORS WHERE RDB$GENERATOR_ID=1';
 Result :=StringReplace(Result,'?',FNameGenerator,[]);
end;

{ TSelectWithParam_ }

constructor TSelectWithParam_.Create(connect: TConnection; t: TTransaction; isParametr : boolean);
var TPB : TReadOnlyTPB;
    i : integer;
begin
 FTransaction :=nil; //       
 inherited Create(isParametr);

//    
 if t = nil then begin
   FIsOutTransaction :=false;
   TPB :=TReadOnlyTPB.Create;
   try
     FTransaction :=TTransaction.Create;
     FTransaction.Start_transaction(connect,TPB);
   finally
    TPB.Free;
   end;
// 
  end else begin
    FIsOutTransaction :=true;
    FTransaction :=t;
  end;
  FConnect :=connect;

//  SQL 
  Prepare(connect,FTransaction,GetStrSelect);
//  
  for i :=0 to GetCountFields-1 do
    if not FRecord_.Set_XSQLVAR(Field[i]) then
      raise EIbaseError.Create('  .    , , .');
// 
  if isParametr then GetParams([]);
end;

destructor TSelectWithParam_.Destroy;
begin
 if IsOpen then Close_cursor;
 if (FTransaction <> nil) and not FIsOutTransaction then begin
   if FTransaction.IsActiveTransaction then FTransaction.Commit;
   FTransaction.Free;
 end;
 FRecord_.Free;
 inherited Destroy;
end;

procedure TSelectWithParam_.GetParams(p: array of TField_);
var i : integer;
begin
 for i :=0 to High(p) do begin
  StrCopy(Parametr[i]^.sqlname,PChar(p[i].NameField));
  Parametr[i]^.sqlname_length :=length(p[i].NameField);
  if p[i].GetType = tfString then
   Parametr[i]^.sqllen :=length(p[i].FValueString);
  p[i].Set_XSQLVAR(Parametr[i]);
 end;
end;

{ TRecordCountRecords }

constructor TRecordCountRecords.Create;
begin
 inherited Create;
 AddField(tfInteger,'count_').SetAlias('count_');
end;

function TRecordCountRecords.GetCount: integer;
begin
 Result :=(GetField('count_') as TFieldInteger).Value;
end;

{ TSelectCountRecords }

constructor TSelectCountRecords.Create(connect: TConnection; table_: string; where_ : string ='');
begin
 FNameTable :=table_;
 FWhere :=where_;
 FRecord_ :=TRecordCountRecords.Create;
 inherited Create(connect,nil);
end;

function TSelectCountRecords.GetRecord: TRecordCountRecords;
begin
 Result :=(FRecord_ as TRecordCountRecords);
end;

function TSelectCountRecords.GetStrSelect: string;
begin
 Result :='SELECT COUNT(*) AS COUNT_ FROM ' + FNameTable;
 if FWhere <> '' then Result :=Result + ' WHERE ' + FWhere;
end;

{ TFieldTimeStamp }

constructor TFieldTimeStamp.Create(NameField: string);
begin
 inherited Create(NameField);
 Value :=0; // 
end;

function TFieldTimeStamp.GetDate: TDateTime;
begin
 Result :=TimestampTo(@FValueTimestamp);
end;

function TFieldTimeStamp.GetType: TTypeField_;
begin
 Result :=tfTimestamp;
end;

procedure TFieldTimeStamp.SetDate(const Value: TDateTime);
begin
 ToTimestamp(@FValueTimestamp,Value); isNull :=Value = 0;
end;

{ TConcurrencyOnlyReadTPB }

constructor TConcurrencyOnlyReadTPB.Create;
 var ParametrTPB : TParametrTPB;
 begin
  inherited Create;
  ParametrTPB :=TParametrTPB.Create;
  try
    ParametrTPB.Access_mode :=tpb_read;            //   
    ParametrTPB.Isolation_level :=tpb_concurrency; //     
    ParametrTPB.Lock_resolution :=tpb_nowait;      //   
    FillTPB(ParametrTPB);
  finally
   ParametrTPB.Free;
  end;
end;

{ TConcurrencyOnlyReadTransaction }

constructor TConcurrencyOnlyReadTransaction.Create(connect: TConnection);
var TPB_ : TConcurrencyOnlyReadTPB;
begin
 inherited Create;
 TPB_ :=TConcurrencyOnlyReadTPB.Create;
 try
   Start_transaction(connect,TPB_);
 finally
   TPB_.Free;
 end;
end;

destructor TConcurrencyOnlyReadTransaction.Destroy;
begin
 if IsActiveTransaction then Commit;
 inherited;
end;

initialization
 SectionAccessToDecimalSeparator :=TCriticalSection.Create;

finalization
 SectionAccessToDecimalSeparator.Free;


end.
